use crate::{prelude::client_pin::{PinUvAuthParam, self}, extensions::cred_protect};

use super::{Error, Request, Response, RelyingParty, Credential};
use fido_common::{credential::public_key, Sha256Hash};
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, Bytes};
use std::borrow::Cow;

#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
#[serde(into = "u8", try_from = "u8")]
pub(super) enum RawSubcommand {
    GetCredsMetadata = 0x01,
    EnumerateRpsBegin = 0x02,
    EnumerateRpsGetNextRp = 0x03,
    EnumerateCredentialsBegin = 0x04,
    EnumerateCredentialsGetNextCredential = 0x05,
    DeleteCredential = 0x06,
    UpdateUserInformation = 0x07,
}

impl From<RawSubcommand> for u8 {
    fn from(val: RawSubcommand) -> Self {
        val as u8
    }
}

impl TryFrom<u8> for RawSubcommand {
    type Error = Error;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Ok(match value {
            0x01 => RawSubcommand::GetCredsMetadata,
            0x02 => RawSubcommand::EnumerateRpsBegin,
            0x03 => RawSubcommand::EnumerateRpsGetNextRp,
            0x04 => RawSubcommand::EnumerateCredentialsBegin,
            0x05 => RawSubcommand::EnumerateCredentialsGetNextCredential,
            0x06 => RawSubcommand::DeleteCredential,
            0x07 => RawSubcommand::UpdateUserInformation,
            _ => return Err(Error::InvalidParameter),
        })
    }
}

#[serde_as]
#[derive(Clone, Serialize, Deserialize)]
pub(super) struct RawSubcommandParams<'a> {
    #[serde(rename = 0x01)]
    pub rp_id_hash: Option<Sha256Hash>,
    #[serde(rename = 0x02)]
    pub credential_id: Option<Cow<'a, public_key::Descriptor>>,
    #[serde(rename = 0x03)]
    pub user: Option<Cow<'a, public_key::UserEntity>>,
}

#[serde_as]
#[derive(Clone, Serialize, Deserialize)]
pub(super) struct RawRequest<'a> {
    #[serde(rename = 0x01)]
    pub subcommand: RawSubcommand,
    #[serde(rename = 0x02, skip_serializing_if = "Option::is_none")]
    pub subcommand_params: Option<RawSubcommandParams<'a>>,
    #[serde(rename = 0x03)]
    pub pin_uv_auth_protocol: client_pin::auth_protocol::Version,
    #[serde_as(as = "Bytes")]
    #[serde(rename = 0x04)]
    pub pin_uv_auth_param: PinUvAuthParam,
}

impl<'a> From<Request<'a>> for RawRequest<'a> {
    fn from(value: Request<'a>) -> Self {
        match value {
            Request::GetCredentialsMetadata {
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            } => RawRequest {
                subcommand: RawSubcommand::GetCredsMetadata,
                subcommand_params: None,
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            },
            Request::EnumerateRPsBegin {
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            } => RawRequest {
                subcommand: RawSubcommand::EnumerateRpsBegin,
                subcommand_params: None,
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            },
            Request::EnumerateRPsGetNextRP => todo!(),
            Request::EnumerateCredentialsBegin {
                relying_party_id_hash,
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            } => RawRequest {
                subcommand: RawSubcommand::EnumerateCredentialsBegin,
                subcommand_params: Some(RawSubcommandParams {
                    rp_id_hash: Some(relying_party_id_hash),
                    credential_id: None,
                    user: None,
                }),
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            },
            Request::EnumerateCredentialsGetNextCredential => todo!(),
            Request::DeleteCredential {
                credential_id,
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            } => RawRequest {
                subcommand: RawSubcommand::DeleteCredential,
                subcommand_params: Some(RawSubcommandParams {
                    rp_id_hash: None,
                    credential_id: Some(credential_id),
                    user: None,
                }),
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            },
            Request::UpdateUserInformation {
                credential_id,
                user,
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            } => RawRequest {
                subcommand: RawSubcommand::UpdateUserInformation,
                subcommand_params: Some(RawSubcommandParams {
                    rp_id_hash: None,
                    credential_id: Some(credential_id),
                    user: Some(user),
                }),
                pin_uv_auth_protocol,
                pin_uv_auth_param,
            },
        }
    }
}

impl<'a> TryFrom<RawRequest<'a>> for Request<'a> {
    type Error = Error;

    fn try_from(value: RawRequest<'a>) -> Result<Self, Self::Error> {
        Ok(match value.subcommand {
            RawSubcommand::GetCredsMetadata => Request::GetCredentialsMetadata {
                pin_uv_auth_protocol: value.pin_uv_auth_protocol,
                pin_uv_auth_param: value.pin_uv_auth_param,
            },
            RawSubcommand::EnumerateRpsBegin => Request::EnumerateRPsBegin {
                pin_uv_auth_protocol: value.pin_uv_auth_protocol,
                pin_uv_auth_param: value.pin_uv_auth_param,
            },
            RawSubcommand::EnumerateRpsGetNextRp => Request::EnumerateRPsGetNextRP,
            RawSubcommand::EnumerateCredentialsBegin => Request::EnumerateCredentialsBegin {
                relying_party_id_hash: value
                    .subcommand_params
                    .ok_or(Error::MissingParameter)?
                    .rp_id_hash
                    .ok_or(Error::MissingParameter)?,
                pin_uv_auth_protocol: value.pin_uv_auth_protocol,
                pin_uv_auth_param: value.pin_uv_auth_param,
            },
            RawSubcommand::EnumerateCredentialsGetNextCredential => {
                Request::EnumerateCredentialsGetNextCredential
            }
            RawSubcommand::DeleteCredential => Request::DeleteCredential {
                credential_id: value
                    .subcommand_params
                    .ok_or(Error::MissingParameter)?
                    .credential_id
                    .ok_or(Error::MissingParameter)?,
                pin_uv_auth_protocol: value.pin_uv_auth_protocol,
                pin_uv_auth_param: value.pin_uv_auth_param,
            },
            RawSubcommand::UpdateUserInformation => {
                let subcommand_params = value.subcommand_params.ok_or(Error::MissingParameter)?;
                Request::UpdateUserInformation {
                    credential_id: subcommand_params
                        .credential_id
                        .ok_or(Error::MissingParameter)?,
                    user: subcommand_params.user.ok_or(Error::MissingParameter)?,
                    pin_uv_auth_protocol: value.pin_uv_auth_protocol,
                    pin_uv_auth_param: value.pin_uv_auth_param,
                }
            }
        })
    }
}

#[derive(Serialize, Deserialize)]
pub(super) struct RawResponse {
    #[serde(rename = 0x01)]
    pub existing_resident_credentials_count: Option<usize>,
    #[serde(rename = 0x02)]
    pub max_possible_remaining_resident_credentials_count: Option<usize>,
    #[serde(rename = 0x03)]
    pub rp: Option<public_key::RelyingPartyEntity>,
    #[serde(rename = 0x04)]
    pub rp_id_hash: Option<Sha256Hash>,
    #[serde(rename = 0x05)]
    pub total_rps: Option<usize>,
    #[serde(rename = 0x06)]
    pub user: Option<public_key::UserEntity>,
    #[serde(rename = 0x07)]
    pub credential_id: Option<public_key::Descriptor>,
    #[serde(rename = 0x08)]
    pub public_key: Option<Vec<u8>>, // TODO: COSE_Key type
    #[serde(rename = 0x09)]
    pub total_credentials: Option<usize>,
    #[serde(rename = 0x0A)]
    pub cred_protect: Option<cred_protect::Policy>,
    #[serde(rename = 0x0B)]
    pub large_blob_key: Option<Vec<u8>>,
    #[serde(rename = 0x0C)]
    pub third_party_payment: Option<bool>,
}

impl From<Response> for RawResponse {
    fn from(value: Response) -> Self {
        match value {
            Response::GetCredentialsMetadata {
                existing_resident_credentials_count,
                max_possible_remaining_resident_credentials_count,
            } => RawResponse {
                existing_resident_credentials_count: Some(existing_resident_credentials_count),
                max_possible_remaining_resident_credentials_count: Some(
                    max_possible_remaining_resident_credentials_count,
                ),
                rp: None,
                rp_id_hash: None,
                total_rps: None,
                user: None,
                credential_id: None,
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::EnumerateRPsBegin {
                relying_party,
                total_relying_parties,
            } => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: Some(relying_party.relying_party),
                rp_id_hash: Some(relying_party.relying_party_id_hash),
                total_rps: Some(total_relying_parties),
                user: None,
                credential_id: None,
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::EnumerateRPsGetNextRP { relying_party } => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: Some(relying_party.relying_party),
                rp_id_hash: Some(relying_party.relying_party_id_hash),
                total_rps: None,
                user: None,
                credential_id: None,
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::EnumerateCredentialsBegin {
                credential,
                total_credentials,
            } => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: None,
                rp_id_hash: None,
                total_rps: None,
                user: Some(credential.user),
                credential_id: Some(credential.credential_id),
                public_key: None,
                total_credentials: Some(total_credentials),
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::EnumerateCredentialsGetNextCredential { credential } => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: None,
                rp_id_hash: None,
                total_rps: None,
                user: Some(credential.user),
                credential_id: Some(credential.credential_id),
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::DeleteCredential => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: None,
                rp_id_hash: None,
                total_rps: None,
                user: None,
                credential_id: None,
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
            Response::UpdateUserInformation => RawResponse {
                existing_resident_credentials_count: None,
                max_possible_remaining_resident_credentials_count: None,
                rp: None,
                rp_id_hash: None,
                total_rps: None,
                user: None,
                credential_id: None,
                public_key: None,
                total_credentials: None,
                cred_protect: None,
                large_blob_key: None,
                third_party_payment: None,
            },
        }
    }
}

impl TryFrom<RawResponse> for Response {
    type Error = Error;

    fn try_from(value: RawResponse) -> Result<Self, Self::Error> {
        // These values are manually checked becaues I guess the client is expected to
        // hold the state of which type of request was made and subsequently
        // which type of response it is expecting, and thus the response does
        // not have a subcommand field.
        //
        // This should be compared against other implementations to see if there is a
        // better way to handle this.
        match value {
            RawResponse {
                existing_resident_credentials_count: Some(existing),
                max_possible_remaining_resident_credentials_count: Some(max),
                ..
            } => Ok(Self::GetCredentialsMetadata {
                existing_resident_credentials_count: existing,
                max_possible_remaining_resident_credentials_count: max,
            }),
            RawResponse {
                rp: Some(rp),
                rp_id_hash: Some(rp_id_hash),
                total_rps: Some(total_rps),
                ..
            } => Ok(Self::EnumerateRPsBegin {
                relying_party: RelyingParty {
                    relying_party: rp,
                    relying_party_id_hash: rp_id_hash,
                },
                total_relying_parties: total_rps,
            }),
            RawResponse {
                rp: Some(rp),
                rp_id_hash: Some(rp_id_hash),
                ..
            } => Ok(Self::EnumerateRPsGetNextRP {
                relying_party: RelyingParty {
                    relying_party: rp,
                    relying_party_id_hash: rp_id_hash,
                },
            }),
            RawResponse {
                user: Some(user),
                credential_id: Some(credential_id),
                public_key: Some(public_key),
                total_credentials: Some(total_credentials),
                cred_protect: Some(cred_protect),
                large_blob_key: Some(large_blob_key),
                ..
            } => Ok(Self::EnumerateCredentialsBegin {
                credential: Credential {
                    user,
                    credential_id,
                    public_key,
                    credential_protection_policy: cred_protect,
                    large_blob_key,
                },
                total_credentials,
            }),
            RawResponse {
                user: Some(user),
                public_key: Some(public_key),
                credential_id: Some(credential_id),
                cred_protect: Some(cred_protect),
                large_blob_key: Some(large_blob_key),
                ..
            } => Ok(Self::EnumerateCredentialsGetNextCredential {
                credential: Credential {
                    user,
                    credential_id,
                    public_key,
                    credential_protection_policy: cred_protect,
                    large_blob_key,
                },
            }),
            // Looks like we have to have the context of what the request was to
            // know the response type when no fields are set...
            _ => todo!(),
        }
    }
}
