extern crate hidapi;

use crate::hid::HidAuthenticator;
use ctap2_proto::prelude::{
    client_pin::auth_protocol::{self, platform::Session},
    credential::public_key,
    *,
};
use rand::{distributions, Rng};
use std::sync::{LazyLock, Mutex};

mod test_client_pin;
mod test_credential_management;

static AUTHENTICATOR: LazyLock<Mutex<Option<HidAuthenticator>>> =
    LazyLock::new(|| Mutex::new(get_authenticator()));

fn init() {
    let _ = env_logger::builder().is_test(true).try_init();
}

fn get_authenticator() -> Option<HidAuthenticator> {
    init(); // done here to ensure init is only run once

    let hidapi = hidapi::HidApi::new().unwrap();
    for info in hidapi.device_list() {
        let Ok(device) = info.open_device(&hidapi) else {
            continue;
        };
        let Ok(authenticator) = device.try_into() else { continue };
        return Some(authenticator);
    }

    None
}

fn get_session(authenticator: &mut HidAuthenticator) -> ctap2_platform::Session {
    let info = authenticator.get_info();
    assert!(info
        .pin_uv_auth_protocols
        .unwrap()
        .contains(&auth_protocol::Version::One));

    let version = auth_protocol::Version::One;

    // Getting Shared Secret K
    let req = client_pin::Request::GetKeyAgreement { version };
    println!("request: {req:#?}");
    let res = authenticator.client_pin(req).unwrap();
    println!("response: {res:#?}");

    let client_pin::Response::GetKeyAgreement { key_agreement } = res else {
        panic!("unexpected response");
    };
    let session = ctap2_platform::Session::initialize(key_agreement).unwrap();

    return session;
}

#[test]
fn test_get_info() {
    let guard = AUTHENTICATOR.lock().unwrap();
    let authenticator = guard.as_ref().unwrap();

    let info = authenticator.get_info();
    println!("deserialized: {:#?}", info);
}

#[test]
fn test_reset() {
    let mut guard = AUTHENTICATOR.lock().unwrap();
    let guarded_authenticator = guard.as_mut().unwrap();

    println!("Remove and re-insert your YubiKey to perform the reset...");

    // Wait for the authenticator to be removed
    while let Some(_) = get_authenticator() {
        std::thread::sleep(std::time::Duration::from_millis(500));
    }

    // Wait for the authenticator to be re-inserted
    let authenticator = loop {
        if let Some(authenticator) = get_authenticator() {
            break authenticator;
        }
        std::thread::sleep(std::time::Duration::from_millis(500));
    };

    *guarded_authenticator = authenticator;

    // Reset the authenticator
    guarded_authenticator.reset().unwrap();
}

#[test]
fn test_make_credential() {
    let mut guard = AUTHENTICATOR.lock().unwrap();
    let authenticator = guard.as_mut().unwrap();

    let info = authenticator.get_info();

    let client_data_hash: Vec<u8> = rand::thread_rng()
        .sample_iter(&distributions::Standard)
        .take(32)
        .collect();

    let user_id = rand::thread_rng()
        .sample_iter(&distributions::Standard)
        .take(32)
        .collect();

    let rp = public_key::RelyingPartyEntity {
        id: "com.example".to_string(),
        name: Some("Example Inc.".into()),
    };

    let user = public_key::UserEntity {
        id: user_id,
        name: Some("example_user".to_string()),
        display_name: Some("Example User".to_string()),
    };

    let pub_key_params: Vec<_> = info.algorithms.unwrap().into_iter().collect();

    let options = [(make::OptionKey::Discoverable, true)].into();

    let req = make::Request::builder()
        .client_data_hash(&client_data_hash.split_array_ref::<32>().0)
        .relying_party(&rp)
        .user(&user)
        .public_key_credential_params(&pub_key_params)
        .options(&options)
        .build();

    println!("request: {req:#?}");
    let response = authenticator.make_credential(req);
    println!("response: {response:#?}");

    let req = get::Request {
        relying_party_id: &rp.id,
        client_data_hash: client_data_hash.as_slice().try_into().unwrap(),
        allow_list: None,
        extensions: None,
        options: None,
        pin_uv_auth_param: None,
        pin_uv_auth_protocol_version: None,
    };

    println!("request: {req:#?}");
    let response = authenticator.get_assertion(req);
    println!("response: {response:#?}");
}
